"""           \_,---;>= Tzefa
    Python Visual Hebrew support by man@zzzen.com
    Based on the Hebrew "Lingua" Perl module by
    Stanislav Malyshev <frodo@sharat.co.il>

    Public functions:
        log2vis_string - Convert a string from "logical" to "visual" Hebrew
        log2vis_text - Convert multiline text (See its docstring for details)
        word_unescape - Fix strange chars MS-Word sometimes uses for '-','"', etc.
"""

####################Copyright stuff################################
# This is A Python adaptation of the Hebrew Lingua Perl lib by
# Stanislav Malyshev <frodo@sharat.co.il>
# I've changed some stuff. I'm not even sure the logic's the same.
# Want it? It's yours.
# Nimrod Kerrett <Man@zzzen.com>
################Original copyright stuff for the Perl Lingua module ##############
#   Logical-Visual Hebrew subroutines
#   Author: Stanislav Malyshev <frodo@sharat.co.il>
#   Date: 02/08/1998
#   Based on the algorithm from the book 'The Unicode Standard, Version 2.0'
#   Redistribution and modification of the code is allowed freely under LGPL terms
##################################################################################

import string
import re

DEBUG=0 ########## debug level (0 is none, 2 is heavy)

ALEF='\340'
TAV='\372'
RE_HEB=re.compile('[%s-%s]' % (ALEF,TAV))

# Direction constants
DIR_NONE=0
DIR_LTR=1
DIR_RTL=2
DIR_NEUTRAL=3
DIR_EMBED=4

DEBUG_DIR={
    DIR_NONE:'',
    DIR_LTR:'l',
    DIR_RTL:'r',
    DIR_NEUTRAL:'n',
    DIR_EMBED:'e'}

def debug_dir(x):
    if DEBUG_DIR.has_key(x):
        return DEBUG_DIR[x]
    return '?'

# Entity type constants
# See Unicode 2.0, page 4-11
STRONG_RTL=0
STRONG_LTR=1
WEAK_EN=2
WEAK_CS=3
NEUTRAL_WS=4
NEUTRAL_ON=5
WEAK_ET=6
WEAK_ES=7

DEBUG_TYPE={
    STRONG_RTL:'R',
    STRONG_LTR:'L',
    WEAK_EN:'EN',
    WEAK_CS:'CS',
    NEUTRAL_WS:'WS',
    NEUTRAL_ON:'N',
    WEAK_ET:'ET',
    WEAK_ES:'ES'}

def debug_type(x):
    if DEBUG_TYPE.has_key(x):
        return DEBUG_TYPE[x]
    return '?'

ENTITY_CHECKS=[
    (RE_HEB,STRONG_RTL),
    (re.compile('[a-zA-Z]'),STRONG_LTR),
    (re.compile('[0-9]'),WEAK_EN),
    (re.compile('[-\$%+\^#]'),WEAK_ET), # Malyshev did 'mulitline' here. Why?
    (re.compile('[./\(\)]'),WEAK_ES), # Malyshev did 'mulitline' here. Why?
    (re.compile('[:,;]'),WEAK_CS),
    (re.compile('\s'),NEUTRAL_WS) ]

def get_type(char):
    for check in ENTITY_CHECKS:
        if check[0].match(char):
            return check[1]
    return NEUTRAL_ON

# This is the list of "mirrored" characters
# See Unicode 2.0, page 4-22
MIRROR={'(':')', ')':'(',
        '[':']', ']':'[',
        '{':'}', '}':'{'}

def log2vis_string(str):
    """ Convert <str> from "logical" to "visual" Hebrew """

    if not RE_HEB.search(str):
        return str # No Heb. Chaval al hazman.
    strlen=len(str)

    ### get entity types    
    str_types=map(get_type,str)
    # a patch to avoid erors when accessing str_types[i+1]
    # (didn't bother Perl, but Python is strict)
    str_types.append(None)

    if DEBUG: print map(debug_type,str_types) 

    ### tweak the weak types
    for i in range(strlen):
        # EN,ES,EN -> EN,EN,EN
        if str_types[i] == WEAK_ES and str_types[i-1] == WEAK_EN and str_types[i+1] == WEAK_EN:
            str_types[i]=WEAK_EN
            continue
        # EN,CS,EN -> EN,EN,EN
        if str_types[i] == WEAK_CS and str_types[i-1] == WEAK_EN and str_types[i+1] == WEAK_EN:
            str_types[i]=WEAK_EN
            continue
        # EN, ET -> EN,EN
        if i>0 and str_types[i-1] == WEAK_EN and str_types[i] == WEAK_ET and str_types[i+1] != STRONG_RTL:
            str_types[i]=WEAK_EN
            continue
        # ET, EN -> EN,EN
        if str_types[i+1] == WEAK_EN and str_types[i] == WEAK_ET:
            str_types[i]=WEAK_EN
            continue
        # otherwise: all weak neutrals become neutral
        if str_types[i] == WEAK_CS or str_types[i] == WEAK_ES or str_types[i] == WEAK_ET:
            str_types[i]=NEUTRAL_ON
            continue

    if DEBUG: print map(debug_type,str_types) 

    ### Compute directions
    levels=[DIR_EMBED]*strlen # default is DIR_EMBED
    next_strong=[NEUTRAL_ON]*strlen
    base=DIR_NONE
    last_strong=0
    for i in range(strlen):
        # first strong character is LTR - all before are LTR
        if str_types[i] == STRONG_LTR:
            levels[i] = DIR_LTR
            if base==DIR_NONE:
                base=DIR_LTR
            for j in range (last_strong,i):
                next_strong[j]=STRONG_LTR
            last_strong=i;
            continue

        # first strong character is RTL - all before are RTL
        if str_types[i] == STRONG_RTL:
            levels[i] = DIR_RTL
            if base==DIR_NONE:
                base=DIR_RTL
            for j in range (last_strong,i):
                next_strong[j]=STRONG_RTL
            last_strong=i
            continue
        
        # directioning neutrals
        if str_types[i] == NEUTRAL_ON or str_types[i] == NEUTRAL_WS:
            # RNR -> RRR
            if str_types[i-1] == STRONG_RTL and str_types[i+1] == STRONG_RTL:
                levels[i]=DIR_RTL
                continue
            # LNL -> LLL
            if str_types[i-1] == STRONG_LTR and str_types[i+1] == STRONG_LTR:
                levels[i]=DIR_LTR
                continue
            # RNL -> REL
            if str_types[i-1] == STRONG_RTL and str_types[i+1] == STRONG_LTR:
               levels[i]=DIR_EMBED
               continue
            # LNR -> LER
            if str_types[i-1] == STRONG_LTR and str_types[i+1] == STRONG_RTL:
               levels[i]=DIR_EMBED
               continue
            # RNW -> RR?
            if str_types[i-1] == STRONG_RTL and str_types[i+1] == WEAK_EN:
                levels[i]=DIR_RTL
                continue
            # LNW -> LL?
            if str_types[i-1] == STRONG_LTR and str_types[i+1] == WEAK_EN:
                levels[i]=DIR_LTR
                continue
            # if basic directionality is RTL : WNL -> ?LL
            if base == DIR_RTL and str_types[i-1] == WEAK_EN and str_types[i+1] == STRONG_LTR:
                levels[i]=DIR_LTR
                continue
            # if basic directionality is RTL : WNR -> ?RR
            if base == DIR_RTL and str_types[i-1] == WEAK_EN and str_types[i+1] == STRONG_RTL:
                levels[i]=DIR_RTL
                continue

        # weak entities get DIR_NEUTRAL
        if str_types[i] == WEAK_EN:
            levels[i]=DIR_NEUTRAL
            continue
    if DEBUG: print map(debug_dir,levels) 

    # compose string
    dir=base # current direction
    cursor=0 # where to put next char
    outstr=''
    nowdir=dir
    for i in range(strlen):
        c=str[i:i+1]
        if levels[i] in (DIR_RTL,DIR_LTR):
            dir=levels[i]
        if levels[i]==DIR_EMBED:
            if dir==DIR_LTR and next_strong[i]==STRONG_RTL:
                dir=DIR_RTL
            elif next_strong[i]==NEUTRAL_ON:
                dir=base
        nowdir=dir
        
        # space between LTR and RTL is moved towards RTL, like
        # abc ABC -> CBA abc
        if levels[i]==DIR_EMBED and dir==DIR_LTR and next_strong[i] == STRONG_RTL:
            nowdir=DIR_RTL

        if levels[i]==DIR_NEUTRAL:
            nowdir=DIR_LTR
        
        if nowdir==DIR_RTL:
            cursor=0
            if MIRROR.has_key(c):
                c=MIRROR[c]
        outstr=outstr[:cursor]+c+outstr[cursor:]
        if nowdir==DIR_LTR:
            cursor=cursor+1
        if DEBUG>1: print "%s:%s:%s [%s]" % (debug_dir(dir),debug_dir(nowdir),cursor,outstr)
    return outstr


WORD_ESCAPE_MAP=None # We "lazy load" it first time it is needed

def word_unescape(str):
    """ Map strange chars MS-Word sometimes uses for '-','`', etc. to non-problematic ASCII """

    global WORD_ESCAPE_MAP
    if not WORD_ESCAPE_MAP: # First time. Should compile patterns
        ### Note: I'm sure this list is not complete.
        ### Please eMail your contributions to man@zzzen.com
        WORD_ESCAPE_MAP=((re.compile("&#8211"),'-'),
                         (re.compile("&#8212"),'-'),
                         (re.compile("&#8222"),'`'),
                         (re.compile("&#8217"),"'"),
                         (re.compile("&#8222"),'"'))
    for mapping in WORD_ESCAPE_MAP:
        str=re.sub(mapping[0],mapping[1],str)
    return str

def log2vis_text(text,line_len=80,before='<NOBR>',after="</NOBR><BR>\n",pad_blanks=0,unescape_word=1):
    """ Convert multilne string <text> from "logical" to "visual" Hebrew.
        Break to <line_len> sized lines of whole words (if possible).
        
    """
    outtext=""
    if unescape_word: text=word_unescape(text)
    for line in string.split(text,'\n'):
        visstr=log2vis_string(line)
        # split visstr to line_len sized segments, right to left
        while visstr:
            split_point=len(visstr)-line_len # take right line_len chars
            ###@@@ print "before " + str(split_point)
            if split_point<=0: # line is short enough
                split_point=0
            else: # find a ' ' after split_point (actually, it's "before" in R2L)
                sp=string.find(visstr,' ',split_point)
                if sp>=0:
                    split_point=sp
                else: # last word is a mosnter (>line_len)
                    sp=string.rfind(visstr,' ',0,split_point) # find ' ' before (i.e. "after") "monster word"
                    if sp>0:
                        split_point=sp # "cut corners" and give whole "monster word"
                    else: # "monster word" is all we have. Take whole string
                        split_point=0
            ###@@@ print "after " + str(split_point)
            visstr,thisline=(string.rstrip(visstr[:split_point]),
                            string.lstrip(visstr[split_point:]))
            if pad_blanks:
                thisline=string.rjust(thisline,line_len)
            outtext=outtext+before+thisline+after
    return outtext

def test(infile,outfile,line_len=80,before='<NOBR>',after="</NOBR><BR>\n",pad_blanks=0,unescape_word=1):
    fin=open(infile,"r")
    fout=open(outfile,"w")
    fout.write(log2vis_text(fin.read(),line_len,before,after,pad_blanks,unescape_word))
    fin.close()
    fout.close()
